# -*- encoding: utf-8 -*-

'''
Author     :Phailin
Datetime   :2023-08-01 09:00:00
E-Mail     :phailin791@hotmail.com
'''

from pydantic import BaseModel
from typing import Generic, TypeVar, List

T = TypeVar("T")


class Page(BaseModel, Generic[T]):
    total: int = 0  # type: ignore
    page: int = 1  # type: ignore
    page_size: int = 10  # type: ignore
    page_data: List[T] = [] # type: ignore

class RenderBase(BaseModel, Generic[T]):
    code: int = 10200  # first version api code,
    data: T = ""
    message: str = ""

class RenderPage(BaseModel, Generic[T]):
    code: int = 10200  # first version api code,
    data: Page[T] = ""
    message: str = ""


# -*- encoding: utf-8 -*-
'''
@Time    :   2023-08-01 09:00:00
@Author  :   phailin791 
@Version :   1.0
@Contact :   phailin791@hotmail.com
'''

from loguru import logger
from typing import Generator
from jose import JWTError, jwt

from fastapi import Depends, HTTPException, status, Request
from fastapi.security import OAuth2PasswordBearer
from sqlalchemy.orm import Session

ALGORITHM = "HS256"
reusable_oauth2 = OAuth2PasswordBearer(tokenUrl=f"/api/v1/login/access-token")

def get_current_user(request: Request, token: str = Depends(reusable_oauth2)):
    # init exception
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    # token_key
    token_key = request.app.container.config.security.secret_key()
    try:
        payload = jwt.decode(token, token_key, algorithms=[ALGORITHM])
        user_id: str = payload.get("sub")
        if user_id is None:
            raise credentials_exception
    except JWTError:
        raise credentials_exception
    user = request.app.container.user.user_repository().get(user_id)
    if user is None:
        raise credentials_exception
    return user


# -*- encoding: utf-8 -*-

'''
Author     :Phailin
Datetime   :2023-08-01 09:00:00
E-Mail     :phailin791@hotmail.com
'''

from suzaku_framework.register_auto.register import RegisterMeta

# global service registry
service_registry = {}

# derector to register service
def register_service(cls):
    return RegisterMeta(service_registry)(cls)


# -*- encoding: utf-8 -*-

'''
Author     :Phailin
Datetime   :2023-08-01 09:00:00
E-Mail     :phailin791@hotmail.com
'''

import traceback
from loguru import logger
from fastapi.responses import JSONResponse
from fastapi import FastAPI, Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from suzaku_framework import custom_exception

def init_exception(app: FastAPI):
    # catch custom Exception
    @app.exception_handler(custom_exception.ParamError)
    async def request_param_exception_handler(request: Request, exc: custom_exception.ParamError):
        """
            custom exception of param exception
        """
        logger.error(f"400_param_exception\n - URL:{request.url}\nHeaders:{request.headers}\n{traceback.format_exc()}")
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content={"code": 10400, "data": "", "message": exc.message},
        )

    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(request: Request, exc: RequestValidationError):
        """
            custom exception of validator exception 
        """
        logger.error(f"400_validator_exception\n - URL:{request.url}\nHeaders:{request.headers}\n{traceback.format_exc()}")
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content=jsonable_encoder({"code": 10400, "data": "", "body": exc.body, "message": exc.errors()}),
        )

    @app.exception_handler(custom_exception.TokenError)
    async def token_exception_handler(request: Request, exc: custom_exception.TokenError):
        """
            custom exception of token exception
        """
        logger.error(f"401_token_exception\n - URL:{request.url}\nHeaders:{request.headers}\n{traceback.format_exc()}")
        return JSONResponse(
            status_code=status.HTTP_401_UNAUTHORIZED,
            content={"code": 10401, "data": None, "message": exc.message},
        )

    @app.exception_handler(custom_exception.NotFound)
    async def not_found_exception_handler(request: Request, exc: custom_exception.NotFound):
        """
            custom exception of 404
        """
        logger.error(f"404_not_found_exception\n - URL:{request.url}\nHeaders:{request.headers}\n{traceback.format_exc()}")
        return JSONResponse(
             status_code=status.HTTP_404_NOT_FOUND,
             content={"code": 10404, "data": "", "message": exc.message}
        )

    @app.exception_handler(Exception)
    async def global_exception_handler(request: Request, exc: Exception):
        """
            global exception
        """
        logger.error(f"500_global_exception\n - URL:{request.url}\nHeaders:{request.headers}\n{traceback.format_exc()}")
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={"code": 10500, "data": "", "message": "Internal Server Error"},
        )


# -*- encoding: utf-8 -*-

'''
Author     :Phailin
Datetime   :2023-08-01 09:00:00
E-Mail     :phailin791@hotmail.com
'''

from typing import Optional, Any
from datetime import datetime
from pydantic import BaseModel, Field, field_serializer

class BaseSchema(BaseModel):
    id: Optional[int] = Field(None, description="id")
    created_at: Optional[datetime] = Field(None, description="创建时间")
    updated_at: Optional[datetime] = Field(None, description="更新时间")
    comment: Optional[str] = Field(None, max_length=128, description="备注")

    @field_serializer("created_at")
    def serialize_created_at(self, created_at: datetime, _info):
        v = None
        if created_at and isinstance(created_at, datetime):
            v = created_at.strftime("%Y-%m-%d %H:%M:%S")
        return v

    @field_serializer("updated_at")
    def serialize_updated_at(self, updated_at: datetime, _info):
        v = None
        if updated_at and isinstance(updated_at, datetime):
            v = updated_at.strftime("%Y-%m-%d %H:%M:%S")
        return v

    class Config:
        from_attributes = True


# -*- encoding: utf-8 -*-

'''
Author     :Phailin
Datetime   :2023-08-01 09:00:00
E-Mail     :phailin791@hotmail.com
'''

import sys
import importlib
from pathlib import Path
from typing import Dict, List

def get_group_file(member_name: str ="endpoint.py"):
    data = []
    # current_dir = Path(__file__).parent
    current_dir = Path.cwd()
    for path in current_dir.glob('**/*.py'):
         if path.is_file() and path.name == member_name:
            if 'src' in path.parts:
                src_index = path.parts.index('src')
                src_subpath = path.parts[src_index:]
                object_path = '.'.join(src_subpath)
                data.append(object_path.replace('.py', ''))
    return data

def auto_load_from_group(module_names:List[str]):
    imported_modules = {}
    for module_name in module_names:
        try:
            # 假设模块文件名和模块名相同，只是扩展名为.py
            module_file = module_name.replace('.', '/') + '.py'
            module = import_module_from_string(module_name, module_file)
            imported_modules[module_name] = module
        except ImportError as e:
            print(f"Failed to import module {module_name}: {e}")
    return imported_modules

def import_module_from_string(module_name: str, path: str):
    # 将相对路径转换为绝对路径
    absolute_path = Path(path).resolve()
    
    # 假设每个模块文件都在其自己的目录中，目录名即为模块名
    module_path = absolute_path.parent.__str__().replace('\\', '/')
    
    # 将模块路径添加到sys.path中
    sys.path.append(module_path)
    
    try:
        # 动态导入模块
        module = importlib.import_module(module_name)
        return module
    finally:
        # 导入完成后，从sys.path中移除模块路径
        sys.path.remove(module_path)



# -*- encoding: utf-8 -*-

'''
Author     :Phailin
Datetime   :2023-08-01 09:00:00
E-Mail     :phailin791@hotmail.com
'''

class ParamError(Exception):
    def __init__(self, message: str="Request Param Error"):
        self.message = message

class TokenError(Exception):
    def __init__(self, message: str="Token Error"):
        self.message = message

class NotFound(Exception):
    def __init__(self, message: str="Not Found"):
        self.message = message


# -*- encoding: utf-8 -*-

'''
Author     :Phailin
Datetime   :2023-08-01 09:00:00
E-Mail     :phailin791@hotmail.com
'''

from datetime import datetime
from pydantic import BaseModel
from sqlalchemy.orm import Session, Query
from sqlalchemy import delete, insert, update
from contextlib import AbstractContextManager
from fastapi.encoders import jsonable_encoder
from typing import Any, Dict, Generic, Optional, Type, TypeVar, Union,Callable, Iterator, List

from suzaku_framework.database import Base

ModelType = TypeVar("ModelType", bound=Base)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class BaseRepository(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, 
        session_factory: Callable[..., AbstractContextManager[Session]],
        model: Type[ModelType]) -> None:
        self.session_factory = session_factory
        self.model = model

    def get(self, id: Any) -> Optional[ModelType]:
        # database
        with self.session_factory() as session:
            db_obj = session.query(self.model).filter(self.model.id == id, self.model.deleted==0).first()
            return db_obj

    def get_all(self) -> Iterator[ModelType]:
        # database
        with self.session_factory() as session:
            return session.query(self.model).filter_by(deleted=0).all()

    def paginate(self, queryset:Query, page: int = 1, per_page: int = 20):
        # parse per_page
        _query_offset = (page - 1) * per_page
        # items
        items = queryset.limit(per_page).offset(_query_offset).all()
        # count
        total = queryset.order_by(None).count()
        return total, items

    def create(self, obj_in: CreateSchemaType, *args, **kwargs) -> ModelType:
        # parse obj_data
        obj_in_data = jsonable_encoder(obj_in)
        # update obj_data
        obj_in_data.update(**kwargs)
        # instance
        db_obj = self.model(**obj_in_data)  # type: ignore
        # database
        with self.session_factory() as session:
            # set common value
            db_obj.created_at = datetime.now()
            db_obj.deleted = False
            # database
            session.add(db_obj)
            session.commit()
            session.refresh(db_obj)
            return db_obj

    def update(self, db_obj: ModelType, obj_in: Union[UpdateSchemaType, Dict[str, Any]], *args, **kwargs) -> ModelType:
        # parse obj_data
        obj_data = jsonable_encoder(db_obj)
        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.dict(exclude_unset=True)
        # update obj_data
        obj_data.update(**kwargs)
        # set business data of obj_data
        for field in obj_data:
            if field in update_data:
                setattr(db_obj, field, update_data[field])
        # database 
        with self.session_factory() as session:
            # set common value
            db_obj.updated_at = datetime.now()
            # database
            session.add(db_obj)
            session.commit()
            session.refresh(db_obj)
            return db_obj

    def removed(self, db_obj: ModelType, *args, **kwargs) -> ModelType:
        with self.session_factory() as session:
            # set common value
            db_obj.updated_at = datetime.now()
            db_obj.deleted = True
            for field in kwargs:
                if field in jsonable_encoder(db_obj).keys():
                    setattr(db_obj, field, kwargs[field])
            # database
            session.add(db_obj)
            session.commit()
            session.refresh(db_obj)
            return db_obj

    def deleted(self, id: int, *args, **kwargs) -> ModelType:
        with self.session_factory() as session:
            obj = session.query(self.model).get(id)
            session.delete(obj)
            session.commit()
            return obj
        
    def batch_insert(self, db_objs: List[ModelType], *args, **kwargs):
        with self.session_factory() as session:
            for chunk in range(0, len(db_objs)+1, 1000):
                session.add_all(db_objs[chunk:chunk+1000])
                session.flush()
            session.commit()

    def batch_update(self, filter_kwargs: Dict, update_kwargs: Dict):
        with self.session_factory() as session:
            stmt = update(self.model)
            if filter_kwargs:
                stmt = stmt.where(**filter_kwargs)
            if update_kwargs:
                stmt = stmt.values(**update_kwargs)
            # sysnc
            stmt = stmt.execution_options(synchronize_session=False)
            # execute
            results =session.execute((stmt))
            # commit
            session.commit()
            # refresh
            session.refresh(results)
            # return
            return results

    def bulk_update(self, db_objs: List[ModelType], *args, **kwargs)-> List[ModelType]:
        with self.session_factory() as session:
            # update
            for chunk in range(0, len(db_objs), 1000):
                db_objs_chunk = db_objs[chunk:chunk+1000]
                session.execute(
                    update(self.model),
                    db_objs_chunk
                )
                session.flush()
            # commit
            session.commit()